open Myarray;;

type gridProbDist = float array array;;
type goalAchieved = Failed of ((int * int) list * float) list 
  | AchievedAverage of float;;

type cell = Full | Empty | Goal;;
type observation = {up: bool; down: bool; left: bool; right: bool};;
type grid = cell array array;;


(* Those that aren't the goal or full *)
let incEmpty = fun total pos ->
  match pos with 
  | Empty -> total + 1
  | Goal -> total
  | Full -> total 

let gridNumCells = fun grid ->
  fold_left  
    (fun x array -> x + fold_left incEmpty 0 array)
    0
    grid 

let lookCell =fun grid (indx, indy) ->
  wget (wget grid indy) indx     

(* Is cell a valid starting point *)
let isCellStarting grid (indx, indy) =
  match lookCell grid (indx ,indy) with
  | Full -> false
  | Empty -> true
  | Goal -> false

let isCellFull grid (indx, indy) =
  match lookCell grid (indx ,indy) with
  | Full -> true 
  | Empty -> false
  | Goal -> false
;;

let gridCellToObs = fun grid (indx, indy) ->
  { up= isCellFull grid (indx, (indy-1)) ;
    down= isCellFull grid (indx, (indy+1));
    left = isCellFull grid ((indx-1), indy);
    right = isCellFull grid ((indx+1), indy)
  }
;;
let mkUniformDist = fun grid ->
  let num = gridNumCells grid in
  let prob = 1.0 /. float_of_int (num) in
  map 
    (map  (fun gridCell -> match gridCell with
    | Empty -> prob
    |_ -> 0.0 ) )
    grid

let myGrid =  [|
    [| Full; Empty; Empty; Full|];
    [| Full; Empty; Goal; Full|]
  |]

let valueHt = fun ()->
  Hashtbl.create 100 
    
let dumbPolicy = fun obs ->
  if obs.up = true then 
    (0, 1)
  else
    (0, -1)

let fixedPolicyGen = fun action -> 
  fun obs -> action;;

(*Need to canoicalize here *)
let canonPos = fun grid (indx, indy) -> 
  let leny = length grid in
  let lenx = length grid.(0) in
  let retx = if indx >= lenx then indx - lenx
  else if indx < 0 then lenx + indx  else indx in
  let rety = if indy >= leny then indy - leny
  else if indy < 0 then leny + indy  else indy in
  (retx,rety)

let detTran = fun grid (indx, indy) action ->
  let potret = canonPos grid (indx + fst action, indy + snd action) in
  match lookCell grid potret with
  | Empty -> potret
  | Goal -> potret
  | Full -> (indx,indy)
      

let detTranPol = fun grid (indx,indy) pol ->
  detTran grid (indx,indy) (pol (gridCellToObs grid (indx,indy) ) )

(* random action-- could check it's a valid action here *)
let rec sampleAction = fun grid ->
  let rand = Random.int 2 in
  let potrand = 2 * (Random.int 2 ) - 1 in 
  if rand = 0 then
    (potrand,0)
  else
    (0,potrand)


      
(* We should sample a state from our probability density of states *)
(* I'll make this crappy for now and uniformly sample states
   instead of using search on the CDF. *)
let rec sampleState = 
  fun grid ->
    let leny = length grid in
    let lenx = length grid.(0) in
    let potposy = Random.int leny in
    let potposx = Random.int lenx in
    if isCellStarting grid (potposx,potposy) then
      (potposx,potposy)
    else
      sampleState grid
;;

let rec mkStateList =  fun grid stateList n ->
  if n = 0 then 
    stateList
  else
    mkStateList grid ( (sampleState grid)::stateList) (n-1) 

let rec mkActionList =  fun grid actionList n ->
  if n = 0 then 
    actionList
  else
    mkActionList grid ( (sampleAction grid)::actionList) (n-1) 

let computeAveVal = fun valHt obs act ->
  let tup = List.assoc (obs,act) valHt in
  fst tup /. float_of_int (snd tup)

let bestAction  = fun valHt obs -> 
(*  let actionList = [(-1,0);(1,0);(0,-1);(0,1)] in*)
  let actionList = [(0,-1);(-1,0);(1,0);(0,1)] in
  let recentValue = List.map (computeAveVal valHt obs) actionList in
  let combineList = List.combine actionList recentValue in
  let (bestact,actval) = List.fold_left (fun (a,b) (c,d) -> if  b >= d then (a,b) else (c,d))
    ((-1,0),-100000.0) combineList in
  bestact
      
;;

let bestActionPolicy = fun valHt ->
  fun obs -> bestAction valHt obs
  

let rewardPos =fun grid (indx, indy) -> 
  match lookCell grid (indx, indy) with
  | Full -> -100000.0
  | Empty -> -1.0
(*  | Goal -> -1.0 Had to change this to match the RL-ness of this version...*)
  | Goal -> 0.0


let rec propagate grid stateList reward policyList = 
  if List.length policyList > 0 then
    let currPos = List.hd stateList in
    if lookCell grid currPos != Goal then (* we're not done *)
      let nextPos = detTranPol grid currPos (List.hd policyList) in
      let currRew = rewardPos grid nextPos in
      propagate grid (nextPos::stateList) (currRew +. reward) (List.tl policyList)
    else
      (stateList, reward)
  else
    (stateList, reward)
;;


let computeActionValue = fun grid (indx, indy) act policyList ->
  let newPolicyList = (fixedPolicyGen act ):: policyList in
  let (stateList,reward) = propagate grid [(indx,indy)] 0.0 newPolicyList in
  let obs = gridCellToObs grid (List.hd (List.rev stateList)) in
  (obs, act, reward)

(*
let initializeTable = fun valHt (obs,act,reward) ->
  ignore(Hashtbl.add valHt (obs,act) (0.0,0))

let installValInTable = fun valHt (obs,act,reward) ->
  let (currentReward,currentNumber) = Hashtbl.find valHt (obs,act) in 
  Hashtbl.replace valHt (obs,act) (currentReward +. reward,currentNumber + 1)

let computeActionValueFun = fun grid policyList ->
  let initialStateList =  mkStateList grid [] 10 in 
  let initialActionList = mkActionList grid [] 10 in
  let compfun  = fun s a -> computeActionValue grid s a policyList in 
  let results = List.map2 compfun initialStateList initialActionList  in
  let valHt = valueHt () in 
  ignore(List.map (fun result -> initializeTable valHt result) results );
  ignore(List.map (fun result -> installValInTable valHt result) results);
  valHt
;;
*)
(*Also contruct assocList purely functional versions *)

let initializeTable = fun valHt (obs,act,reward) ->
  ((obs,act),(0.0,0))::valHt

let installValInTable = fun valHt (obs,act,reward) ->
  let (currentReward,currentNumber) = List.assoc (obs,act) valHt in
  (  (obs,act),(currentReward +. reward ,currentNumber + 1) )::valHt

let rec installInitial = fun valHt resultsList ->
  if resultsList == [] then
    valHt
  else
    installInitial (initializeTable valHt (List.hd resultsList) ) (List.tl resultsList)


let rec installResults = fun valHt resultsList ->
  if resultsList = [] then
    valHt
  else
    installResults (installValInTable valHt (List.hd resultsList)) (List.tl resultsList)

(*policy pretty-printer*)
let ppPolicy = fun grid policy->
  let leny = length grid in
  let lenx = length grid.(0) in
  let stateList = ref [] in

  for j=0 to leny-1 do
    for i=0 to lenx-1  do
      if isCellStarting grid (i,j) = true then
	match policy (gridCellToObs grid (i,j)) with
	| (-1,0) -> Printf.printf "<"
	| (1,0) -> Printf.printf ">"
	| (0,1) -> Printf.printf "/"
	| (0,-1) -> Printf.printf "^"
	| _ -> Printf.printf "ERROR!!!"
      else if isCellFull grid (i,j) = true then
	Printf.printf "X" 
      else
	Printf.printf "O"
    done;
      Printf.printf "\n"
  done
    
let ppPolicies = fun grid policyList ->
  let f = fun p -> ppPolicy grid p; Printf.printf "\n" in
  List.iter f policyList


let mkStateListComplete = fun grid ->
  let leny = length grid in
  let lenx = length grid.(0) in
  let stateList = ref [] in
  for i=0 to lenx -1 do
    for j=0 to leny-1 do
      if isCellStarting grid (i, j) == true then 
	ignore(stateList := [(i,j);(i,j);(i,j);(i,j)]@(!stateList))
    done
  done;
  !stateList


let mkStateListComplete1x = fun grid ->
  let leny = length grid in
  let lenx = length grid.(0) in
  let stateList = ref [] in
  for i=0 to lenx -1 do
    for j=0 to leny-1 do
      if isCellStarting grid (i, j) == true then 
	ignore(stateList := [(i,j)]@(!stateList))
    done
  done;
  !stateList

let mkActionListComplete = fun stateList ->
  let len = List.length stateList / 4 in
  let actList = ref [] in
  for i=0 to len-1 do
    ignore(actList := [(-1,0);(1,0);(0,-1);(0,1)]@(!actList) )
  done;
  !actList
    

let computeActionValueRanFun = fun grid policyList n->
(*
  let initialStateList =  mkStateListComplete grid in 
  let initialActionList = mkActionListComplete initialStateList in 
*)
  let initialStateList =  mkStateList grid [] n in 
  let initialActionList = mkActionList grid [] n in 
  let compfun  = fun s a -> computeActionValue grid s a policyList in 
  let results = List.map2 compfun initialStateList initialActionList  in
  let valHt = installInitial [] results in
  installResults valHt results


let computeActionValueFun = fun grid policyList ->
  let initialStateList =  mkStateListComplete grid in 
  let initialActionList = mkActionListComplete initialStateList in 
  let compfun  = fun s a -> computeActionValue grid s a policyList in 
  let results = List.map2 compfun initialStateList initialActionList  in
  let valHt = installInitial [] results in
  installResults valHt results 
;;
 
let rec computeNStepPolicy = fun grid n policyList ->
  if n = 0 then
    policyList
  else
    computeNStepPolicy grid (n-1) 
      ((bestActionPolicy 
	  (computeActionValueFun grid policyList))
       ::policyList)
;;

let rec computeNStepPolicyRan = fun grid n policyList nsamples->
  if n = 0 then
    policyList
  else
    computeNStepPolicyRan grid (n-1) 
      ((bestActionPolicy 
	  (computeActionValueRanFun grid policyList nsamples))
       ::policyList) nsamples
;;


let isGoal = fun grid (indx,indy) -> 
  match lookCell grid (indx ,indy) with
  | Full -> false
  | Empty ->false 
  | Goal -> true

let achievesGoal = fun grid (trajList, weightedReward)  ->
  List.fold_left (||) false   (List.map (isGoal grid) trajList);;


let computeWeightedReward = fun grid weightGrid policyList (indx,indy) ->
  let (stateList,reward) = propagate grid [(indx,indy)] 0.0 policyList in
  let weight = lookCell weightGrid (indx,indy) in
  (stateList,reward *. weight)
;;    

let averageReward = fun grid weightGrid policyList ->
  let initialStateList =  mkStateListComplete1x grid in 
  let allPaths = List.map (computeWeightedReward grid weightGrid policyList) initialStateList in
  List.fold_left 
    (fun x y -> ( fst y , 
		  (snd x) +. (snd y) ) 
    )
    ([(0,0)], 0.0)  allPaths 
;;
  
let allAchieveGoal = fun grid weightGrid policyList ->
  let initialStateList =  mkStateListComplete1x grid in 
  let allPaths = List.map (computeWeightedReward grid weightGrid policyList) initialStateList in
  let didAchieve = List.fold_left (&&) true 
      (List.map (achievesGoal grid) allPaths) in
    match didAchieve with
    | true -> (AchievedAverage (snd
	  (List.fold_left 
	  (fun x y -> ( fst y , 
			(snd x) +. (snd y) ) 
	  )
	  ([(0,0)], 0.0)  allPaths)) )
    | false -> Failed (List.filter (fun x -> not (achievesGoal grid x)) allPaths)



